from uuid import uuid4
import string
import itertools


# File Names

def get_uuid_filename(size: int, extension: str) -> list:
    return [f"{uuid4()}{extension or ''}" for val in range(size)]


def get_numeric_filename(size: int, extension: str) -> list:
    format_length = len(f"{size}")
    file_names = [f"{val:0{format_length}}{extension or ''}" for val in range(size)]
    return file_names


def get_alphabetic_filename(size: int, extension: str) -> list:
    repeat_value = None
    for exp in range(100):
        if size <= 26 ** exp:
            repeat_value = exp
            break
    print(f"Using : {repeat_value}")
    alphabet_tuple = itertools.product(string.ascii_lowercase, repeat=repeat_value)
    file_names = [f"{''.join(val)}{extension or ''}" for val in list(itertools.islice(alphabet_tuple, size))]
    return file_names


def get_filename_from_url(url_list: list) -> list:
    file_list = []
    for url in url_list:
        file_name = url.split("?")[0].split("/")[-1]
        file_list.append((file_name, url))
    return file_list


# Reading Files


def get_urls(file_path: str) -> list:
    url_list = []

    with open(file_path, "r") as fp:
        for url in fp:  # type: str
            url_list.append(url.strip())

    return url_list



import argparse
from time import time
from .paralleldownload_module import parallel_download
from .misc import info, version_info, simple_example, detailed_example



from .__init__ import VERSION, CORE_COUNT


def start():
    parser = argparse.ArgumentParser(
        usage="%(prog)s url_file [save_directory] [options]",
        formatter_class=argparse.RawTextHelpFormatter,
        description="Parallel Fast Downloader",
        epilog=simple_example,
        add_help=False,
    )

    # TODO
    # https://docs.python.org/3/library/argparse.html#nargs
    parser.add_argument(
        "url_file",
        help="Text file containing the URLs of the files to be downloaded. One URL per file",
    )

    parser.add_argument(
        "save_directory",
        help="The directory to be used to save the downloaded files. [default: %(default)s (Current Directory)]",
        nargs="?",
        default="."
    )

    parser.add_argument(
        "-p",
        "--process",
        help=f"Parallel Process to use when downloading the files. [default: %(default)s (CPU Core count in the machine)]",
        type=int,
        default=CORE_COUNT
    )

    parser.add_argument(
        "-e",
        "--extension",
        help="Manually specify the Extension to use.",
    )

    file_name_group = parser.add_mutually_exclusive_group()

    file_name_group.add_argument(
        "-u",
        "--uuid",
        help="Use UUID string as the file name",
        action='store_true',
    )

    file_name_group.add_argument(
        "-n",
        "--number",
        help="Use sequential Numbers as the file name",
        action='store_true',
    )

    file_name_group.add_argument(
        "-a",
        "--alphabet",
        help="Use sequential Alphabets as the file name",
        action='store_true',
    )



    misc_parser = parser.add_argument_group('help', 'Help, Logs and Info Options')

    misc_parser.add_argument(
        "-s",
        "--silent",
        help="Don't print Verbose Logs. Print only essential information",
        action='store_true',
    )

    misc_parser.add_argument(
        "-d",
        "--debug",
        help="Print detailed information. Helpful when debugging issues.",
        action='store_true',
    )

    misc_parser.add_argument(
        "-h",
        "--help",
        action="help",
        help="Show this help message and exit"
    )

    misc_parser.add_argument(
        "-v",
        "--version",
        help="Prints Version information of the Package",
        action="version",
        version=version_info,
    )

    misc_parser.add_argument(
        "-i",
        "--info",
        action='version',
        help="Prints Information about the Package",
        version=info,
    )

    misc_parser.add_argument(
        "-eg",
        "--example",
        action='version',
        help="Prints more Example which helps to use this tool",
        version=detailed_example,
    )

    print("Parsing arguments")
    args = parser.parse_args()


    start_time = time()
    status = parallel_download(args)
    end_time = time()

    print(f"The Operation completed in {end_time - start_time :.5f} Seconds.")




from .validate import validate_input_file, validate_save_directory, validate_process, get_valid_extension, \
    divide_into_nested_list, status_code_desc
from .files import get_urls, get_uuid_filename, get_numeric_filename, get_alphabetic_filename

from pathlib import Path
from multiprocessing import Pool
import requests
import json


def parallel_download(args):
    url_file = args.url_file
    save_directory = args.save_directory
    process = args.process
    uuid = args.uuid
    number = args.number
    alphabet = args.alphabet
    silent = args.silent
    debug = args.debug

    if debug:
        print_args(args)

    print("Checking options")

    extension = get_valid_extension(args.extension)

    file_exists, file_readable = validate_input_file(url_file)
    dir_exists, dir_writeable = validate_save_directory(save_directory)
    process_ok = validate_process(process)

    status = None

    if not all([file_exists, file_readable, dir_exists, dir_writeable, process_ok]):
        print("Options invalid")
        print_errors(args,
                     file_exists=file_exists, file_readable=file_readable,
                     dir_exists=dir_exists, dir_writeable=dir_writeable,
                     process_ok=process_ok
                     )
    else:
        print("Options are valid")
        print("Preparing download")
        url_list = get_urls(url_file)
        save_path = Path(save_directory).absolute().as_posix()
        rename_if_possible = True
        download_nodes = []

        if not url_list:
            print_errors(args, file_empty=bool(url_list))
        else:

            urls_size = len(url_list)
            file_name_list = []
            if uuid:
                file_name_list = get_uuid_filename(urls_size, extension)
                rename_if_possible = False
            elif number:
                file_name_list = get_numeric_filename(urls_size, extension)
                rename_if_possible = False
            elif alphabet:
                file_name_list = get_alphabetic_filename(urls_size, extension)
                rename_if_possible = False
            else:
                file_name_list = get_uuid_filename(urls_size, extension)

            for url, file in zip(url_list, file_name_list):
                download_nodes.append({
                    "url": url,
                    "file": file,
                    "dir": save_path,
                    "flags": (rename_if_possible, silent, debug),
                })

            args_list = divide_into_nested_list(download_nodes, process)

            print(f"Distributing work to {process} Process")
            status = download_using_multiprocessing(args_list)
            print_failed_urls(status)

    return status


def print_args(args):
    try:
        args_dict = args.__dict__
        print(json.dumps(args_dict, indent=2))
    except:
        print(args_dict)


def print_failed_urls(failed_nested_list):
    for process_list in failed_nested_list:  # type: list
        for failed_node in process_list:  # type: dict

            url = failed_node['url']
            rename, silent, debug = failed_node['flags']

            status_code = f"{failed_node.get('status_code', 'Unknown')}"
            status_code_category = status_code_desc.get(f"{status_code[:1]}xx", "Unknown")
            status_code_detail = status_code_desc.get(status_code, "Unknown")
            status_code_description = f"The URL {url} failed due to {status_code_category}, The Status code is {status_code} - {status_code_detail}"

            failed_node.pop('file', '')
            failed_node.pop('dir', '')
            failed_node.pop('flags', '')

            if debug:
                try:
                    failed_node['header'] = dict(failed_node['header'])
                    failed_node['body'] = failed_node['body'].decode()
                    failed_node['status_code_description'] = status_code_description
                    print(json.dumps(failed_node, indent=2))
                except:
                    print(status_code_description, failed_node)

            else:
                print(status_code_description)


def print_errors(args, **kwargs):
    print(args, kwargs)

    error_message = {
        "file_exists": f"The File ({args.url_file}) does not exists. Please provide a valid file name / file path",
        "file_readable": f"The File ({args.url_file}) is not readable. Please provide a valid permission before using",
        "dir_exists": f"The Directory / Folder ({args.url_file}) does not exists. Please provide a valid path",
        "dir_writeable": f"The Directory / Folder ({args.url_file}) is not writeable. Please provide a valid permission before using",
        "process_ok": f"The Process mentioned ({args.process}) does not fall in the range 0 < process < 120",
        "file_empty": f"The File ({args.url_file}) is Empty"
    }

    for key, val in kwargs.items():
        if not val:
            print(error_message.get(key))


def download_using_multiprocessing(args_list: list):
    status_list = []
    process_count = len(args_list)
    with Pool(process_count) as process:
        dist_work = process.imap_unordered(download, args_list)
        for status in dist_work:
            status_list.append(status)
    return status_list





def download(download_nodes: list) -> list:
    failed_nodes = []
    # directory remains same for all urls.
    directory = download_nodes[0]['dir']
    for nodes in download_nodes:

        url, file_name = nodes['url'], nodes['file']
        rename, silent, debug = nodes['flags']

        response = requests.get(url, allow_redirects=True, stream=True)

        if response and response.status_code == 200:
            if rename:
                file_name = f"{file_name}"

            full_file_path = f"{directory}/{file_name}"

            with open(full_file_path, "wb") as fp:
                fp.write(response.content)

            if not silent:
                print(f"Download Successful for : {file_name}")
        else:
            nodes['status_code'] = response.status_code

            if debug:
                nodes['header'] = response.headers
                nodes['body'] = response.content

            failed_nodes.append(nodes)
        # end if

    return failed_nodes


def get_file_name_from_response(name):
    pass



import os
from pathlib import Path


def validate_input_file(path: str) -> tuple:
    file_path = Path(path)
    file_exists = False
    file_readable = False
    if file_path.exists() and file_path.is_file():
        file_exists = True

    if os.access(path, os.R_OK):
        file_readable = True

    return file_exists, file_readable


def validate_save_directory(path: str) -> tuple:
    directory_path = Path(path)
    directory_exists = False
    directory_writeable = False
    if directory_path.exists() and directory_path.is_dir():
        directory_exists = True

    if os.access(path, os.W_OK | os.X_OK):
        directory_writeable = True

    return directory_exists, directory_writeable


def validate_process(process: int) -> bool:
    if 0 < process < 120:
        return True
    else:
        return False


def get_valid_extension(extension: str) -> str:
    if extension:
        # Taking only first 20 characters of the extension string.
        extension = extension[:20].strip()
        if extension[0] != ".":
            valid_extension = f".{extension}"
        else:
            valid_extension = extension
    else:
        valid_extension = None

    return valid_extension


def divide_into_nested_list(total_list: list, split_count: int) -> list:
    k = split_count
    n = len(total_list)
    nested_list = [total_list[i * (n // k) + min(i, n % k):(i + 1) * (n // k) + min(i + 1, n % k)] for i in range(k)]
    return nested_list



status_code_desc = {
        "1xx": "Informational",
        "100": "Continue",
        "101": "Switching Protocols",
        "102": "Processing (WebDAV)",
        "2xx": "Success",
        "200": "OK",
        "201": "Created",
        "202": "Accepted",
        "203": "Non-Authoritative Information",
        "204": "No Content",
        "205": "Reset Content",
        "206": "Partial Content",
        "207": "Multi-Status (WebDAV)",
        "208": "Already Reported (WebDAV)",
        "226": "IM Used",
        "3xx": "Redirection",
        "300": "Multiple Choices",
        "301": "Moved Permanently",
        "302": "Found",
        "303": "See Other",
        "304": "Not Modified",
        "305": "Use Proxy",
        "306": "(Unused)",
        "307": "Temporary Redirect",
        "308": "Permanent Redirect (experimental)",
        "4xx": "Client Error",
        "400": "Bad Request",
        "401": "Unauthorized",
        "402": "Payment Required",
        "403": "Forbidden",
        "404": "Not Found",
        "405": "Method Not Allowed",
        "406": "Not Acceptable",
        "407": "Proxy Authentication Required",
        "408": "Request Timeout",
        "409": "Conflict",
        "410": "Gone",
        "411": "Length Required",
        "412": "Precondition Failed",
        "413": "Request Entity Too Large",
        "414": "Request-URI Too Long",
        "415": "Unsupported Media Type",
        "416": "Requested Range Not Satisfiable",
        "417": "Expectation Failed",
        "418": "I'm a teapot (RFC 2324)",
        "420": "Enhance Your Calm (Twitter)",
        "422": "Unprocessable Entity (WebDAV)",
        "423": "Locked (WebDAV)",
        "424": "Failed Dependency (WebDAV)",
        "425": "Reserved for WebDAV",
        "426": "Upgrade Required",
        "428": "Precondition Required",
        "429": "Too Many Requests",
        "431": "Request Header Fields Too Large",
        "444": "No Response (Nginx)",
        "449": "Retry With (Microsoft)",
        "450": "Blocked by Windows Parental Controls (Microsoft)",
        "451": "Unavailable For Legal Reasons",
        "499": "Client Closed Request (Nginx)",
        "5xx": "Server Error",
        "500": "Internal Server Error",
        "501": "Not Implemented",
        "502": "Bad Gateway",
        "503": "Service Unavailable",
        "504": "Gateway Timeout",
        "505": "HTTP Version Not Supported",
        "506": "Variant Also Negotiates (Experimental)",
        "507": "Insufficient Storage (WebDAV)",
        "508": "Loop Detected (WebDAV)",
        "509": "Bandwidth Limit Exceeded (Apache)",
        "510": "Not Extended",
        "511": "Network Authentication Required",
        "598": "Network read timeout error",
        "599": "Network connect timeout error",
    }




# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['paralleldownload']

package_data = \
{'': ['*']}

install_requires = \
['requests>=2.5,<3.0']

entry_points = \
{'console_scripts': ['paralleldownload = '
                     'paralleldownload.paralleldownloadcli:start',
                     'pfd = paralleldownload.paralleldownloadcli:start']}

setup_kwargs = {
    'name': 'paralleldownload',
    'version': '0.5.2',
    'description': 'Parallel Fast Downloader',
    'long_description': '# Paralleldownload\n#### Download large number of files extremely fast.\n\n\nA Python Package for Fast Parallel Download of Multiple Files.\nIt\'s Simple, easy and extremely fast since it uses, all the cores in your CPU to spin separate process for parallel download.\nParalleldownload uses `requests` as its only dependency, which is almost always present in all python environment.\n\n\n\n![PyPI - Python Version](https://img.shields.io/pypi/pyversions/paralleldownload?style=for-the-badge)\n\n### Python Package Index Badges\n\n[![PyPI](https://img.shields.io/pypi/v/paralleldownload?style=for-the-badge&color=gree&logo=pypi)](https://pypi.org/project/batchimage/)\n![PyPI - Downloads](https://img.shields.io/pypi/dm/paralleldownload?label=Downloads&style=for-the-badge)\n![PyPI - Status](https://img.shields.io/pypi/status/paralleldownload?label=Status&style=for-the-badge)\n![PyPI - Format](https://img.shields.io/pypi/format/paralleldownload?label=Format&style=for-the-badge)\n\n\n### Github Badges\n\n![GitHub last commit](https://img.shields.io/github/last-commit/insumanth/paralleldownload?style=for-the-badge)\n![GitHub commit activity](https://img.shields.io/github/commit-activity/y/insumanth/paralleldownload?style=for-the-badge)\n![GitHub code size in bytes](https://img.shields.io/github/languages/code-size/insumanth/paralleldownload?style=for-the-badge)\n![Lines of code](https://img.shields.io/tokei/lines/github/insumanth/paralleldownload?style=for-the-badge)\n\n\n\n------------------\n## Python Package Index Install \n\n```\npip install paralleldownload\n```\n-------------------\n\n### Usage:\n\n**Basic example:**\n```shell\n$ paralleldownload input_url_file.txt\n```\n* It downloads the files using the urls in the file. Each url in a line.\n* The downloaded files is stored in current directory.\n* Uses number of process equal to number of CPU Cores in the machine.\n\n**Getting help, info, version and example:**\n```shell\n$ paralleldownload [-h | -i | -v | -eg]\n```\n* These options will just print text and exits.\n  * `-h ` Prints the help message.\n  * `-i ` Prints information aboout the package.\n  * `-v ` Prints current version of the package.\n  * `-eg` Prints few example of how to use the this cli.\n\nSpecify save directory:\n```shell\n$ paralleldownload input_url_file.txt downloads_directory\n```\n* The downloaded files is stored in `downloads_directory` directory.\n* You can provide absolute or relative paths for both url file and save directory.\n\n**Use N Processes**:\n```shell\n$ paralleldownload input_url_file.txt -p 17\n```\n* Uses 17 processes to download files.\n* Default is equal to the number of cpu core count in the machine.\n\n\n**File names:**\n\n```shell\n$ paralleldownload input_url_file.txt [-u | -n | -a]\n```\n\n* By default, it searches for the file name in the response. If found, it will use this name. Else it will use a uuid string as file name. [Yet to be implemented]\n  * `-u` All downloaded files will be named as uuid strings `[eg: 5b71113f-43be-40f5-b267-9b93919196aa.jpg]`\n  * `-n` All downloaded files will be named as sequential numbers `[eg: 017.jpg]`\n  * `-a` All downloaded files will be named as sequential lowercase english alphabets `[eg: exy.jpg]`\n* If extension is needed, they have to be manually provided.\n\n\n**Specify Extension**\n```shell\n$ paralleldownload input_url_file.txt -e png\n```\n* Uses the provided extension in file names.\n* `.` (dot) is optional.\n* If extension is needed, they must be provided when using `[-u | -n | -a]` .\n\n\n\n------------------\n\n#### Description\n\nDo you want to download thousands of files at once but can\'t wait for sequential download? \n\nToday\'s machines have multiple CPU cores.\nMost entry level machines have 4 Cores while higher end machines have around 8 Cores, \nSome desktop processor even have 16 - 32 Cores.\nBut using just one Core for downloading files is not the best approach if you have hundreds or thousands of files to download.\n\nThe rapid shift towards cloud technologies provide massive processing power, GigaBit network and faster writes to disk.\nBy properly making use of this processing power, bandwidth, memory and IO, we can make our life a bit easier.\n\nparalleldownload is one of such package.\nIt is a cli tool used to download thousands of files in short time.\nIt achieves it by spinning seperate process per CPU core and downloading parallely.\n\n\nImagine a Machine with 24 Core CPU and Gigabit Network.\nThe process to download 1000 files, and it takes 1 second to fetch each file,\n\n##### Sequential download\n\nPython Interpreter does the following steps.\n\nMake request  ➜ open a file ➜ save the content to the file\n\nAfter it is done, Python Interpreter repeats the steps 1000 times sequentially.\n\n\n\nIt takes `1000 files` `x` `1 Second` = `1000 Seconds` or `~17 Minutes` to download all files.\n\n`➜➜➜ .. ➜➜➜ .. ➜➜➜ .. ➜➜➜ .. ➜➜➜ .. ➜➜➜ .. ➜➜➜ .. ➜➜➜`\n \n##### Parallel download\n\n28 Python Interpreter is spun up in all of 28 cores (one per core by default).\n\nEach Python Interpreter does the following steps.\n\nMake request  ➜ open a file ➜ save the content to the file\n\nAfter it is done, Each Python Interpreter repeats the steps 42 times (1000 files / 24 process).\n\n```\n➜➜➜ .. ➜➜➜ .. ➜➜➜ .. ➜➜➜ .. ➜➜➜ .. ➜➜➜ .. ➜➜➜ .. ➜➜➜\n➜➜➜ .. ➜➜➜ .. ➜➜➜ .. ➜➜➜ .. ➜➜➜ .. ➜➜➜ .. ➜➜➜ .. ➜➜➜\n[.. 21 more processes ..]\n➜➜➜ .. ➜➜➜ .. ➜➜➜ .. ➜➜➜ .. ➜➜➜ .. ➜➜➜ .. ➜➜➜ .. ➜➜➜\n```\n\nIt takes `(1000 files` `/` `24 Processes)` `x` `1 Second` = `42 Seconds` or `~1 Minutes` to download all files.\n\n**Note:** This is just a logical comparison.\nThe time is not a real world example, many factors affects the download speeed.\nThere will be a slight overhead in setting up the process and collecting the work.\nSo, it will not download all in under a minute.\nThe Overhead should be negligable with higher input.\n\n\n\n\n\n-----------------\n![Made With Python](https://forthebadge.com/images/badges/made-with-python.svg)\n![forthebadge](https://forthebadge.com/images/badges/built-with-love.svg)\n\n<!--\n<img src="https://forthebadge.com/images/badges/works-on-my-machine.svg" alt="works-on-my-machine" width="500" height="100">\n-->\n',
    'author': 'Sumanth',
    'author_email': 'sumanthreddystar@gmail.com',
    'maintainer': 'Sumanth',
    'maintainer_email': 'sumanthreddystar@gmail.com',
    'url': 'https://pypi.org/project/paralleldownload/',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)




from .__init__ import VERSION, CORE_COUNT

info = """
 
    Info
    -----
        Name    : Parallel Downloader
        Version : 0.1.0 
 
    """

version_info = f"""
 
 Parallel Downloader Version : {VERSION}
 
"""

simple_example = f"""
 
    Simple Example
    --------------
    $ paralleldownload input_url_file.txt 
   
   * It downloads the files using the urls in the file. Each url in a line.
   * Uses number of process equal to number of CPU Cores in the machine [You have {CORE_COUNT} CPU Cores]
 
"""

detailed_example = f"""
{simple_example}
"""
